# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
from builtins import map, filter, str
from operator import itemgetter

import mistune
import requests
from bs4 import BeautifulSoup
from future.utils import iteritems
from html2text import html2text
from six.moves.urllib.parse import urljoin

from anpy.utils import extract_datetime

AN_BASE_URL = 'http://www.assemblee-nationale.fr'


class InvalidResponseException(Exception):
    pass


class Dossier(object):
    def __init__(self, url=None, senat_url=None, title=None,
                 legislature=None, procedure=None, steps=None):
        self.url = url
        self.senat_url = senat_url
        self.title = title
        self.legislature = legislature
        self.procedure = procedure
        self.steps = steps or []

    @staticmethod
    def download_and_build(url):
        resp = requests.get(url)
        if resp.status_code >= 400:
            raise InvalidResponseException('%s: %d' % (url, resp.status_code))
        return DossierParser(url, resp.content).parse()

    def to_dict(self):
        return {
            'url': self.url,
            'senat_url': self.senat_url,
            'title': self.title,
            'legislature': self.legislature,
            'procedure': self.procedure,
            'steps': self.steps
        }


class ProcedureParlementaire(object):
    PJL = 'PJL'
    PPL = 'PPL'


class LegislativeStepType(object):
    AN_PREMIERE_LECTURE = 'AN_PREMIERE_LECTURE'
    SENAT_PREMIERE_LECTURE = 'SENAT_PREMIERE_LECTURE'
    AN_DEUXIEME_LECTURE = 'AN_DEUXIEME_LECTURE'
    SENAT_DEUXIEME_LECTURE = 'SENAT_DEUXIEME_LECTURE'
    AN_TROISIEME_LECTURE = 'AN_TROISIEME_LECTURE'
    SENAT_TROISIEME_LECTURE = 'SENAT_TROISIEME_LECTURE'
    CMP = 'CMP'
    CONSEIL_CONSTIT = 'CONSEIL_CONSTITUT'
    AN_NOUVELLE_LECTURE = 'AN_NOUVELLE_LECTURE'
    SENAT_NOUVELLE_LECTURE = 'SENAT_NOUVELLE_LECTURE'
    AN_LECTURE_DEFINITIVE = 'AN_LECTURE_DEFINITIVE'


class DecisionStatus(object):
    ADOPTE = 'ADOPTE'
    REJETE = 'REJETE'
    MODIFIE = 'MODIFIE'


class LegislativeActType(object):
    ETUDE_IMPACT = 'ETUDE_IMPACT'
    AVIS_CONSEIL_ETAT = 'AVIS_CONSEIL_DETAT'
    PROCEDURE_ACCELEREE = 'PROCEDURE_ACCELEREE'
    REUNION_COMMISSION = 'REUNION_COMMISSION'
    DISCUSSION_COMMISSION = 'DISCUSSION_COMMISSION'
    TEXTE_COMMISSION = 'TEXTE_COMMISSION'
    DEPOT_RAPPORT = 'DEPOT_RAPPORT'
    SAISIE_COM_AVIS = 'SAISIE_COM_AVIS'
    NOMINATION_RAPPORTEURS = 'NOMINATION_RAPPORTEURS'
    DISCUSSION_SEANCE_PUBLIQUE = 'DISCUSSION_SEANCE_PUBLIQUE'
    DECISION = 'DECISION'
    DEPOT_INITIATIVE = 'DEPOT_INITIATIVE'
    SAISIE_COM_FOND = 'SAISIE_COM_FOND'
    PROMULGATION = 'PROMULGATION'
    SAISINE_CONSEIL_CONSTIT = 'SAISINE_CONSEIL_CONSTIT'


class DossierParser(object):
    def __init__(self, url, html):
        self.url = url
        self.html = html
        self.soup = BeautifulSoup(clean_html(html), 'html5lib')

    def parse(self):
        return Dossier(
            url=self.url,
            title=self.parse_title(),
            legislature=self.parse_legislature(self.url),
            procedure=self.parse_procedure(),
            steps=self.parse_steps(),
            senat_url=self.parse_senat_url())

    def parse_title(self):
        return self.soup.strong.text if self.soup.strong else None

    def parse_procedure(self):
        re_procedure = re.compile('(projet de loi|proposition de loi)', re.I)

        def match_procedure(a):
            return re_procedure.match(a.text)

        def get_procedure(a):
            proc = match_procedure(a).group(1).lower()
            if proc == 'projet de loi':
                return ProcedureParlementaire.PJL
            elif proc == 'proposition de loi':
                return ProcedureParlementaire.PPL

        return next(
            map(get_procedure,
                filter(match_procedure, self.soup.find_all('a'))), None)

    def parse_senat_url(self):
        re_url = re.compile('http://www.senat.fr/dossier-legislatif')
        re_url_alt = re.compile('http://www.senat.fr/dossierleg/')

        return next(
            filter(lambda url: re_url.match(url) or re_url_alt.match(url),
                   map(lambda a: a['href'], self.soup.find_all('a'))), None)

    def parse_legislature(self, url=None):
        re_legislature = re.compile(r'/(\d{2})/')

        def match_legislature(a):
            return re_legislature.match(a['href'])

        def get_matched_group(a):
            return match_legislature(a).group(1) \
                if match_legislature(a) else None

        if url and 'assemblee-nationale' in url:
            legislature = url.split('.fr/')[1].split('/')[0]
            if legislature:
                return legislature

        return next(map(get_matched_group,
                        filter(match_legislature, self.soup.find_all('a'))))

    def parse_steps(self):
        tree = self.build_step_tree()
        return tree.extract_data()

    def build_step_tree(self):
        root_node = DossierNode()
        current_node = root_node

        for element in filter(filter_dossier_element,
                              self.soup.find_all(['p', 'h2'])):
            new_node_class = BaseNode.match_node_class(element)

            if new_node_class:
                parent = current_node.get_relevant_parent(new_node_class)
                next_current_node = new_node_class(parent=parent)
                parent.add_child(next_current_node)
                current_node = next_current_node

            current_node.add_element(element)

        return root_node


def clean_html(html):
    soup = BeautifulSoup(html, 'html5lib')
    if soup.body.header:
        soup.body.header.extract()
    md_text = html2text(str(soup.body), bodywidth=0)
    return mistune.markdown(md_text).replace('<br>\n', '</p>\n<p>')


def filter_dossier_element(element):
    return element.text.strip() and \
           not element.text.startswith('_') and \
           not element.text.startswith('Accueil')


class BaseNode(object):
    def __init__(self, parent=None):
        self.parent = parent
        self.elements = []
        self.children = []

    @classmethod
    def match(cls, html):
        raise NotImplementedError

    @classmethod
    def match_node_class(cls, element):
        for class_ in [LegislativeStepNode,
                       DiscussionSeancePubliqueNode,
                       DepotLoiNode,
                       ProcedureAccelereeNode,
                       AvisConseilEtatNode,
                       EtudeImpactNode,
                       DepotTexteCommissionNode,
                       RapportNode,
                       DecisionNode]:
            if class_.match(element):
                return class_

    def get_relevant_parent(self, node_class):
        return self

    def add_child(self, node):
        self.children.append(node)

    def add_element(self, element):
        self.elements.append(element)

    def extract_data(self):
        raise NotImplementedError

    def __repr__(self):
        return '<%s [%s]>' % (
            self.__class__.__name__,
            self.elements[0].__repr__() if self.elements else '')


class DossierNode(BaseNode):
    def extract_data(self):
        return [step.extract_data() for step in self.children]


class LegislativeStepNode(BaseNode):
    steps_re = {
        LegislativeStepType.AN_PREMIERE_LECTURE:
            re.compile('^assemblée nationale - 1 ?ère lecture',
                       re.I | re.UNICODE),
        LegislativeStepType.SENAT_PREMIERE_LECTURE:
            re.compile('^sénat - 1 ?ère lecture',
                       re.I | re.UNICODE),
        LegislativeStepType.AN_DEUXIEME_LECTURE:
            re.compile('^assemblée nationale - 2 ?e lecture',
                       re.I | re.UNICODE),
        LegislativeStepType.SENAT_DEUXIEME_LECTURE:
            re.compile('^sénat - 2 ?e lecture',
                       re.I | re.UNICODE),
        LegislativeStepType.AN_TROISIEME_LECTURE:
            re.compile('^assemblée nationale - 3 ?e lecture',
                       re.I | re.UNICODE),
        LegislativeStepType.SENAT_TROISIEME_LECTURE:
            re.compile('^sénat - 3 ?e lecture',
                       re.I | re.UNICODE),
        LegislativeStepType.AN_NOUVELLE_LECTURE:
            re.compile('^assemblée nationale - nouvelle lecture',
                       re.I | re.UNICODE),
        LegislativeStepType.SENAT_NOUVELLE_LECTURE:
            re.compile('^sénat - nouvelle lecture',
                       re.I | re.UNICODE),
        LegislativeStepType.AN_LECTURE_DEFINITIVE:
            re.compile('^assemblée nationale - lecture définitive',
                       re.I | re.UNICODE),
        LegislativeStepType.CONSEIL_CONSTIT:
            re.compile('^conseil constitutionnel',
                       re.I | re.UNICODE),
        LegislativeStepType.CMP:
            re.compile(r'^commission mixte paritaire \((accord|désaccord)?\)$',
                       re.I | re.UNICODE)
    }

    @classmethod
    def match(cls, html):
        return any(map(lambda r: r.match(html.text), cls.steps_re.values()))

    def get_relevant_parent(self, node_class):
        return self.parent if node_class == LegislativeStepNode else self

    def extract_data(self):
        data = {
            'type': self.extract_type(),
            'acts': [act for child in self.children
                     for act in (child.extract_data() or [])]
        }

        if data['type'] == LegislativeStepType.CMP and self.elements:
            if 'Accord' in self.elements[0].text:
                data['status'] = 'ACCORD'
            if 'Désaccord' in self.elements[0].text:
                data['status'] = 'DESACCORD'

        return data

    def extract_type(self):
        if not self.elements:
            return

        return next(
            map(itemgetter(0),
                filter(lambda item: item[1].match(self.elements[0].text),
                       iteritems(self.steps_re))))


class LegislativeActNode(BaseNode):
    rule = None

    @classmethod
    def match(cls, html):
        return cls.rule.match(html.text) if cls.rule else False

    def get_relevant_parent(self, node_class):
        if self.parent.parent:
            return self.parent if issubclass(node_class, LegislativeActNode)\
                else self.parent.parent
        return self.parent

    def extract_data(self):
        raise NotImplementedError


class ProcedureAccelereeNode(LegislativeActNode):
    rule = re.compile('^le gouvernement a engagé la procédure accélérée',
                      re.I | re.UNICODE)

    def extract_data(self):
        if not self.elements:
            return

        matched_dates = re.findall(r' le (\d+\s?\w* \w+ \d{4})',
                                   self.elements[0].text,
                                   re.I | re.UNICODE)

        return [{
            'type': LegislativeActType.PROCEDURE_ACCELEREE,
            'date': extract_datetime(matched_dates[0])
            if matched_dates else None
        }]


class EtudeImpactNode(LegislativeActNode):
    rule = re.compile(r'^etude d\'impact', re.I | re.UNICODE)

    def extract_data(self):
        return [{
            'type': LegislativeActType.ETUDE_IMPACT,
            'url': urljoin(AN_BASE_URL, self.elements[0].a['href'])
        }]


class DepotLoiNode(LegislativeActNode):
    rule = re.compile('^(projet de loi|proposition de loi).+déposée? le',
                      re.I | re.UNICODE)

    @classmethod
    def match(cls, html):
        return cls.rule.match(html.text) and html.a if cls.rule else False

    def extract_data(self):
        if not self.elements:
            return

        return [{
            'type': LegislativeActType.DEPOT_INITIATIVE,
            'url': self.extract_url(),
            'date': self.extract_date()
        }]

    def extract_date(self):
        matched_dates = re.findall(r' déposée? le (\d+ \w+ \d{4})',
                                   self.elements[0].text, re.UNICODE)

        return extract_datetime(matched_dates[0]) if matched_dates else None

    def extract_url(self):
        return urljoin(AN_BASE_URL, self.elements[0].a['href']) \
            if self.elements else None

    def extract_legislature(self):
        link = self.elements[0].a
        matched = re.match(r'/(\d{2})/', link['href'])

        if matched:
            return matched.group(1)

    def extract_procedure(self):
        if self.elements and 'Projet de loi' in self.elements[0].text:
            return ProcedureParlementaire.PJL
        elif self.elements and 'Proposition de loi' in self.elements[0].text:
            return ProcedureParlementaire.PPL


class DepotTexteCommissionNode(LegislativeActNode):
    rule = re.compile('^texte de la commission.+(déposée?|mis en ligne) le',
                      re.I | re.UNICODE)

    @classmethod
    def match(cls, html):
        return cls.rule.match(html.text) and html.a if cls.rule else False

    def extract_data(self):
        if not self.elements:
            return

        return [{
            'type': LegislativeActType.TEXTE_COMMISSION,
            'url': self.extract_url(),
            'date': self.extract_date()
        }]

    def extract_date(self):
        matched_dates = re.findall(
            r'(?:déposée?|mis en ligne) le (\d+ \w+ \d{4})',
            self.elements[0].text, re.UNICODE)

        return extract_datetime(matched_dates[0]) if matched_dates else None

    def extract_url(self):
        return urljoin(AN_BASE_URL, self.elements[0].a['href']) \
            if self.elements else None


class RapportNode(LegislativeActNode):
    rule = re.compile('^rapport.+déposée? le',
                      re.I | re.UNICODE)

    @classmethod
    def match(cls, html):
        return cls.rule.match(html.text) and html.a if cls.rule else False

    def extract_data(self):
        if not self.elements:
            return

        return [{
            'type': LegislativeActType.DEPOT_RAPPORT,
            'url': self.extract_url(),
            'date': self.extract_date()
        }]

    def extract_date(self):
        matched_dates = re.findall(r' déposée? le (\d+ \w+ \d{4})',
                                   self.elements[0].text, re.UNICODE)

        return extract_datetime(matched_dates[0]) if matched_dates else None

    def extract_url(self):
        return urljoin(AN_BASE_URL, self.elements[0].a['href']) \
            if self.elements else None


class DecisionNode(LegislativeActNode):
    rule = re.compile('^(projet de loi|proposition de loi)(?!.*déposé)',
                      re.I | re.UNICODE)

    def extract_data(self):
        status = None
        if 'adopté' in self.elements[0].text:
            status = DecisionStatus.ADOPTE
        elif 'modifié' in self.elements[0].text:
            status = DecisionStatus.MODIFIE
        elif 'rejeté' in self.elements[0].text:
            status = DecisionStatus.REJETE

        matched_dates = re.findall(r' le (\d+\s?\w* \w+ \d{4})',
                                   self.elements[0].text,
                                   re.I | re.UNICODE)

        return [{
            'type': LegislativeActType.DECISION,
            'status': status,
            'date': extract_datetime(matched_dates[0])
            if matched_dates else None,
            'url': urljoin(AN_BASE_URL, self.elements[0].a['href'])
            if self.elements[0].a else None
        }]


class DiscussionSeancePubliqueNode(LegislativeActNode):
    rule = re.compile('^discussion en séance publique', re.I | re.UNICODE)

    def extract_data(self):
        return self.extract_seances_from_first_element() \
            if self.seances_in_first_element() \
            else self.extract_seances_from_all_elements()

    def seances_in_first_element(self):
        return 'au cours des séances' in self.elements[0].text

    def extract_seances_from_first_element(self):
        return [{'url': urljoin(AN_BASE_URL, a['href']),
                 'type': LegislativeActType.DISCUSSION_SEANCE_PUBLIQUE}
                for a in self.elements[0].find_all('a')]

    def extract_seances_from_all_elements(self):
        def extract_seance(element):
            return {
                'url': urljoin(AN_BASE_URL, element.a['href'])
                if element.a else None,
                'date': extract_datetime(element.text.split('séance du ')[1]),
                'type': LegislativeActType.DISCUSSION_SEANCE_PUBLIQUE
            }

        def filter_seance(element):
            return 'séance du ' in element.text and \
                   'Scrutin' not in element.text

        return list(map(extract_seance, filter(filter_seance, self.elements)))


class AvisConseilEtatNode(LegislativeActNode):
    rule = re.compile(r'^avis du conseil d\'État', re.I | re.UNICODE)

    def extract_data(self):
        if not self.elements[0].a:
            return
        return [{
            'type': LegislativeActType.AVIS_CONSEIL_ETAT,
            'url': urljoin(AN_BASE_URL, self.elements[0].a['href'])
        }]



# -*- coding: utf-8 -*-

#           _   _ _______     __
#     /\   | \ | |  __ \ \   / /
#    /  \  |  \| | |__) \ \_/ /
#   / /\ \ | . ` |  ___/ \   /
#  / ____ \| |\  | |      | |
# /_/    \_\_| \_|_|      |_|

# Set default logging handler to avoid "No handler found" warnings.
import logging

__title__ = 'anpy'
__version__ = '0.1.25'
__author__ = 'Regards Citoyens'
__license__ = 'MIT'

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



# re-write of the dosleg parser to have
# the same output as senapy

import sys
import re
from urllib.parse import urljoin

import dateparser
from bs4 import BeautifulSoup

from lawfactory_utils.urls import clean_url, download, parse_national_assembly_url, AN_OLD_URL_TEMPLATE

from anpy.dossier_from_opendata import parse as opendata_parse


def format_date(date):
    parsed = dateparser.parse(date, languages=['fr'])
    return parsed.strftime("%Y-%m-%d")


def find_promulgation_date(line):
    """
    >>> find_promulgation_date("Loi  nº 2010-383 du 16 avril 2010 autorisant l'approbation de l'accord entre...")
    '2010-04-16'
    """
    line = line.split(' du ')[1]
    return format_date(re.search(r"(\d\d? \w\w\w+ \d\d\d\d)", line).group(1))


def find_senat_url(data):
    if not data['steps']:
        return
    senat_text_url = [step['source_url'] for step in data['steps'] if step.get('source_url') and 'senat.fr' in step.get('source_url')]
    for url in senat_text_url:
        html = download(url).text
        soup = BeautifulSoup(html, 'lxml')
        for a in soup.select('#primary a'):
            href = urljoin(url, a.attrs.get('href', ''))
            if 'dossier-legislatif/' in href or 'dossierleg/' in href:
                return clean_url(href)


def download_historic_dosleg(url):
    resp = download(url)

    if '/dyn/' in resp.url:
        # fallback to backed-up doslegs when the redirect is forced
        legislature, slug = parse_national_assembly_url(url)
        display_url = AN_OLD_URL_TEMPLATE.format(legislature=legislature, slug=slug)
        download_url = 'https://raw.githubusercontent.com/regardscitoyens/archive-AN-doslegs/master/archive/' \
            + display_url.split('.fr/')[1]
        resp = download(download_url)
        resp.url = display_url

    resp.encoding = 'Windows-1252'
    return resp


def merge_previous_works_an(older_dos, dos):
    # remove promulgation step
    if older_dos['steps'] and older_dos['steps'][-1].get('stage') == 'promulgation':
        older_dos['steps'] = older_dos['steps'][:-1]

    for i in range(len(older_dos['steps'])):
        if older_dos['steps'][-i].get('source_url') == dos['steps'][0].get('source_url'):
            dos['steps'] = older_dos['steps'][:-i] + dos['steps']
            break
    else:
        dos['steps'] = older_dos['steps'] + dos['steps']

    if 'url_dossier_senat' in older_dos and 'url_dossier_senat' not in dos:
        dos['url_dossier_senat'] = older_dos['url_dossier_senat']

    return dos


def historic_doslegs_parse(html, url_an=None, logfile=sys.stderr, nth_dos_in_page=0, parse_previous_works=True, parse_next_works=True):
    """
    Parse an AN dosleg like http://www.assemblee-nationale.fr/13/dossiers/accord_Montenegro_mobilite_jeunes.asp

    nth_dos_in_page, parse_previous_works and parse_next_works are for internal logic
    """

    data = {
        'url_dossier_assemblee': clean_url(url_an),
        'urgence': False,
    }

    def log_error(*error):
        print('## ERROR ###', *error, file=logfile)

    def log_warning(*error):
        print('## WARNING ###', *error, file=logfile)

    soup = BeautifulSoup(html, 'lxml')

    legislature, slug = parse_national_assembly_url(data['url_dossier_assemblee'])
    data['assemblee_slug'] = slug
    if legislature:
        data['assemblee_legislature'] = legislature
    else:  # strange link (old dosleg)
        log_error('NO LEGISLATURE IN AN LINK: ' + data['url_dossier_assemblee'])
    data['assemblee_id'] = '%s-%s' % (data.get('assemblee_legislature', ''), data['assemblee_slug'])

    data['steps'] = []
    curr_institution = 'assemblee'
    curr_stage = '1ère lecture'
    last_section = None  # Travaux des commissions/Discussion en séance publique
    last_step_index = 0
    travaux_prep_already = False
    another_dosleg_inside = None
    predicted_next_step = None  # For unfinished projects, we try to catch the next step
    previous_works = None
    url_jo = None

    html_lines = html.split('\n')
    for i, line in enumerate(html_lines):
        def parse_line():
            return BeautifulSoup(line, 'lxml')

        def line_text():
            return parse_line().text.strip()

        def get_last_step():
            if len(data['steps']) > 0:
                return data['steps'][-1]
            return {}

        if '<COMMENTAIRE>' in line or '<table border="1"' in line:
            continue

        if '<font face="ARIAL" size="3" color="#000080">' in line:
            data['long_title'] = line_text()
        if '<br><b><font color="#000099">Travaux des commissions</font></b><br>' in line:
            last_section = line_text()
        if '<p align="center"><b><font color="#000080">Travaux préparatoires</font></b><br>' in line:
            if travaux_prep_already:
                if parse_next_works and not nth_dos_in_page:
                    log_warning('FOUND ANOTHER DOSLEG INSIDE THE DOSLEG')
                    another_dosleg_inside = '\n'.join(html.split('\n')[last_step_index + 1:])
                if not nth_dos_in_page:
                    break
                travaux_prep_already = False
            else:
                travaux_prep_already = True
        if not parse_next_works and travaux_prep_already and nth_dos_in_page:
            continue

        # Senat 1ère lecture, CMP, ...
        if '<font color="#000099" size="2" face="Arial">' in line:
            text = line_text()
            last_section = None
            if 'Dossier en ligne sur le site du Sénat' in text:
                data['url_dossier_senat'] = clean_url(parse_line().select(
                    'a')[-1].attrs['href'])
                text = text.replace(
                    '(Dossier en ligne sur le site du Sénat)', '')
            if 'Sénat' in text:
                curr_institution = 'senat'
            elif 'Assemblée nationale' in text:
                curr_institution = 'assemblee'
            elif 'Commission Mixte Paritaire' in text or 'Lecture texte CMP' in text:
                curr_institution = 'CMP'
                curr_stage = 'CMP'
            elif 'Conseil Constitutionnel' in text:
                curr_institution = 'conseil constitutionnel'
                curr_stage = 'constitutionnalité'
            elif 'Congrès du Parlement' in text:
                curr_institution = 'congrès'
                curr_stage = 'congrès'

            if '1ère lecture' in text:
                curr_stage = '1ère lecture'
            elif '2e lecture' in text:
                curr_stage = '2ème lecture'
            elif 'Nouvelle lecture' in text:
                curr_stage = 'nouv. lect.'
            elif 'Lecture définitive' in text:
                curr_stage = 'l. définitive'
            if not curr_stage:
                curr_stage = text.split('-')[-1].strip().lower()

            if curr_stage == "création de la commission d'enquête":
                log_warning('COMMISSION D\'ENQUETE')
                return None

        if '>Proposition de résolution européenne<' in line:
            log_warning('PROPOSITION DE RESOLUTION EUROPEENE')
            return None

        if '>Accès aux Travaux préparatoires' in line and not previous_works:
            previous_works = clean_url(urljoin(url_an, parse_line().find('a').attrs['href']))

        curr_step = None
        # conseil. consti. has no step but we should get the link
        no_step_but_good_link = False
        if 'Rapport portant également sur les propositions' in line:
            continue
        elif re.search(r'<a[^>]* href=[^>]*>(projet de loi|proposition de loi|proposition de résolution)', line, re.I):
            curr_step = 'depot'
            if curr_stage == 'CMP':
                continue
        elif ">Texte de la commission" in line or '/ta-commission/' in line:
            curr_step = 'commission'

        elif '/ta/' in line or '/leg/tas' in line:
            if get_last_step().get('stage') != curr_stage:
                curr_step = 'depot'
                if curr_stage == 'CMP':
                    curr_step = 'commission'
            else:
                curr_step = 'hemicycle'
        elif ('/rapports/' in line or '/rap/' in line) and last_section and 'commissions' in last_section:
            if get_last_step().get('step') == 'commission':
                # log_warning('DOUBLE COMMISSION LINE: %s' % line)
                continue
            curr_step = 'commission'

        elif 'www.conseil-constitutionnel.fr/decision/' in line:
            no_step_but_good_link = True

        # no commissions for l. définitive
        if curr_stage == 'l. définitive' and curr_step == 'commission':
            continue

        if curr_step or no_step_but_good_link:
            # if same step previously, replace or not the url
            if get_last_step().get('step') == curr_step:
                # log_warning('DOUBLE STEP: %s' % line)
                # remove last step since we prefer text links instead of reports links
                # TODO: add report link as bonus_url
                last_url = get_last_step().get('source_url')
                if not last_url or ('/rapports/' in last_url or '/rap/' in last_url):
                    data['steps'] = data['steps'][:-1]
                # looks like the last url was already a text, let's assume it's a multi-depot
                else:
                    # multi-depot if not CMP
                    # TODO: re-order multi depot
                    if curr_institution == 'senat' and curr_stage != 'CMP':
                        curr_step = 'depot'

            links = [a.attrs.get('href') for a in parse_line().select('a')]
            links = [
                href for href in links if href and 'fiches_id' not in href and '/senateur/' not in href and 'javascript:' not in href]
            if not links:
                log_error('NO LINK IN LINE: %s' % (line,))
                continue
            urls_raps = []
            urls_others = []
            for href in links:
                if '/rap/' in href or '/rapports/' in href:
                    urls_raps.append(href)
                else:
                    urls_others.append(href)

            cmp_commission_other_url = None
            if len(urls_others) > 0:
                url = urls_others[0]
                # CMP commission should produce two texts, one for each institution
                if curr_step == 'commission' and curr_stage == 'CMP' and len(urls_others) > 1:
                    cmp_commission_other_url = clean_url(urljoin(url_an, urls_others[1]))
            else:
                url = urls_raps[0]

            url = clean_url(urljoin(url_an, url))

            real_institution = curr_institution
            if curr_stage == 'CMP' and curr_step == 'hemicycle':
                if 'assemblee-nationale.fr' in url:
                    real_institution = 'assemblee'
                elif 'senat.fr' in url:
                    real_institution = 'senat'

            step = {
                'institution': real_institution,
                'stage': curr_stage,
                'source_url': url,
            }

            if curr_step:
                step['step'] = curr_step

            if cmp_commission_other_url:
                step['cmp_commission_other_url'] = cmp_commission_other_url

            # try to detect a date
            for test_line in (line, html_lines[i-1]):
                test_line = test_line.replace('1<sup>er</sup>', '1')
                date_match = re.search(r'(déposée? le|adoptée? .*? le|modifiée? .*?|rejetée? .*?)\s*(\d\d? \w\w\w+ \d\d\d\d)', test_line, re.I)
                if date_match:
                    step['date'] = format_date(date_match.group(2))
                else:
                    date_match = re.search(r'(mis en ligne le)\s*(\d\d? \w\w\w+ \d\d\d\d)', test_line, re.I)
                    if date_match:
                        step['date'] = format_date(date_match.group(2))
                if 'date' in step and 'beginning' not in data:
                    data['beginning'] = step['date']
            data['steps'].append(step)
            predicted_next_step = None
            last_step_index = i

        if 'publiée au Journal Officiel' in line and not url_jo:
            links = [clean_url(a.attrs['href']) for a in parse_line().select('a') if 'legifrance' in a.attrs.get('href', '')]
            if not links:
                log_error('NO GOOD LINK IN LINE: %s' % (line,))
                continue
            url_jo = links[-1]

        if 'Le Gouvernement a engagé la procédure accélérée' in line or 'engagement de la procédure accélérée' in line:
            data['urgence'] = True

        # Next step prediction via small clues
        # TODO: this could be done via last_section (we parse two times the same thing)
        # TODO: this fails for CMP hemicycle senat
        if curr_stage != 'CMP':
            if '>Discussion en séance publique<' in line:
                predicted_next_step = {
                    'institution': curr_institution,
                    'stage': curr_stage,
                    'step': 'hemicycle',
                }
            elif '>Travaux des commissions<' in line:
                predicted_next_step = {
                    'institution': curr_institution,
                    'stage': curr_stage,
                    'step': 'commission',
                }

    metas = {}
    for meta in soup.select('meta'):
        if 'name' in meta.attrs:
            metas[meta.attrs['name']] = meta.attrs['content']

    if not url_jo:
        url_jo = metas.get('LIEN_LOI_PROMULGUEE')

    if url_jo:
        data['url_jo'] = clean_url(url_jo)
        promulgation_step = {
            'institution': 'gouvernement',
            'stage': 'promulgation',
            'source_url': data['url_jo'],
        }
        if metas.get('LOI_PROMULGUEE'):
            data['end'] = find_promulgation_date(metas.get('LOI_PROMULGUEE'))
            promulgation_step['date'] = data['end']
        data['steps'].append(promulgation_step)
    # add predicted next step for unfinished projects
    elif predicted_next_step:
        data['steps'].append(predicted_next_step)

    if 'url_dossier_senat' not in data or 'dossier-legislatif' not in data['url_dossier_senat']:
        senat_url = find_senat_url(data)
        if senat_url:
            data['url_dossier_senat'] = senat_url

    # append previous works if there are some
    if previous_works and parse_previous_works:
        log_warning('MERGING %s WITH PREVIOUS WORKS %s' % (url_an, previous_works))
        resp = download_historic_dosleg(previous_works)
        prev_data = historic_doslegs_parse(
            resp.text, previous_works,
            logfile=logfile,
            nth_dos_in_page=nth_dos_in_page, parse_next_works=False)
        if prev_data:
            prev_data = prev_data[nth_dos_in_page] if len(prev_data) > 1 else prev_data[0]
            data = merge_previous_works_an(prev_data, data)
        else:
            log_warning('INVALID PREVIOUS WORKS', previous_works)

    # is this part of a dosleg previous works ?
    next_legislature = data['assemblee_legislature'] + 1 if 'assemblee_legislature' in data else 9999
    if parse_next_works and next_legislature < 15:
        #  TODO: parse 15th legislature from open data if it exists
        resp = download_historic_dosleg(url_an.replace('/%d/' % data['assemblee_legislature'], '/%d/' % (data['assemblee_legislature'] + 1)))
        if resp.status_code == 200:
            recent_data = historic_doslegs_parse(
                resp.text, resp.url,
                logfile=logfile,
                nth_dos_in_page=nth_dos_in_page, parse_previous_works=False)
            if recent_data:
                log_warning('FOUND MORE RECENT WORKS', resp.url)
                recent_data = recent_data[nth_dos_in_page] if len(recent_data) > 1 else recent_data[0]
                data = merge_previous_works_an(data, recent_data)

    if another_dosleg_inside:
        others = historic_doslegs_parse(another_dosleg_inside, url_an, logfile=logfile, nth_dos_in_page=nth_dos_in_page+1)
        if others:
            return [data] + others
    return [data]


def parse(url, logfile=sys.stderr, cached_opendata_an={}):
    url = clean_url(url)

    if '/dyn/' in url:
        parsed = opendata_parse(url, logfile=logfile, cached_opendata_an=cached_opendata_an)
        if parsed:
            return [parsed]
        print('WARNING: NOT FOUND IN OPEN-DATA', file=logfile)

    resp = download_historic_dosleg(url)
    if resp.status_code != 200:
        print('WARNING: NOT FOUND IN HISTORIC DOSLEGS', file=logfile)
        return []
    return historic_doslegs_parse(resp.text, resp.url, logfile=logfile)


"""
Cas non-gérés (anciens dossiers):
- renvois en commision: http://www.assemblee-nationale.fr/14/dossiers/interdiction_prescription_acquisitive_voies_rurales.asp
- senat ppl manquant: http://www.assemblee-nationale.fr/13/dossiers/comite_poids_mesures.asp
"""



# -*- coding: utf-8 -*-

import logging
import re
import time

import attr
import requests
from bs4 import BeautifulSoup, NavigableString, Comment

from anpy.utils import convert_camelcase_to_underscore

LOGGER = logging.getLogger(__name__)


def parse_amendements_summary(url, json_response):
    """
    json schema :
    {
      infoGenerales: {
        nb_resultats, debut, nb_docs
      },
      data_table: 'id|numInit|titreDossierLegislatif|urlDossierLegislatif|'
                  'instance|numAmend|urlAmend|designationArticle|'
                  'designationAlinea|dateDepot|signataires|sort'
    }

    NB : the json response does not contain the dispositif and expose, that's
    why we call it "amendement's summary"
    """

    amendements = []
    fields = [convert_camelcase_to_underscore(field) for field in
              json_response['infoGenerales']['description_schema'].split('|')]

    for row in json_response['data_table']:
        values = row.split('|')
        amd = AmendementSummary(**dict(zip(fields, values)))
        amd.legislature = re.search(r'www.assemblee-nationale.fr/(\d+)/',
                                    amd.url_amend).groups()[0]
        amendements.append(amd)

    return AmendementSearchResult(**{
        'url': url,
        'total_count': json_response['infoGenerales']['nb_resultats'],
        'start': json_response['infoGenerales']['debut'],
        'size': json_response['infoGenerales']['nb_docs'],
        'results': amendements
    })


def parse_amendement(url, html_response):
    soup = BeautifulSoup(html_response, "html5lib")

    meta_names = [
        'NUM_AMTXT', 'NUM_AMEND', 'AMEND_PARENT', 'URL_DOSSIER', 'NUM_INIT',
        'ETAPE', 'DELIBERATION', 'TITRE_INIT', 'NUM_PARTIE',
        'DESIGNATION_ARTICLE', 'URL_DIVISION', 'DESIGNATION_ALINEA', 'MISSION',
        'AUTEURS', 'AUTEUR_ID', 'GROUPE_ID', 'COSIGNATAIRES_ID', 'SEANCE',
        'SORT', 'DATE_BADAGE', 'DATE_SORT', 'ORDRE_TEXTE', 'CODE', 'REFCODE',
        'LEGISLATURE',
    ]

    kwargs = dict((meta_name.lower(), clean_text(soup.find(
        'meta', attrs={'name': meta_name})['content']))
        for meta_name in meta_names)
    kwargs['auteurs'] = kwargs['auteurs'].replace(u'\xa0', ' ')
    kwargs['dispositif'] = clean_text(remove_inline_css_and_invalid_tags(
        soup.find('dispositif')))
    kwargs['expose'] = clean_text(remove_inline_css_and_invalid_tags(
        soup.find('expose')))
    kwargs['url'] = url

    return Amendement(**kwargs)


class AmendementSearchService(object):
    def __init__(self):
        self.base_url = "http://www2.assemblee-nationale.fr/recherche/query_amendements"  # noqa
        self.default_params = {
            'texteRecherche': None,
            'numAmend': None,
            'idArticle': None,
            'idAuteur': None,
            'idDossierLegislatif': None,
            'idExamen': None,
            'idExamens': None,
            'periodeParlementaire': None,
            'dateDebut': None,
            'dateFin': None,
            'rows': 100,
            'start': None,
            'sort': None,
            'format': 'html',
            'tri': 'ordreTexteasc',
            'typeRes': 'liste',
            'typeDocument': 'amendement',
        }

    def get(self, **kwargs):
        """
        :param texteRecherche:
        :param numAmend:
        :param idArticle:
        :param idAuteur:
        :param idDossierLegislatif:
        :param idExamen:
        :param idExamens:
        :param periodeParlementaire:
        :param dateDebut:
        :param dateFin:
        :param rows:
        :param start:
        :param sort:
        """
        params = self.default_params.copy()
        params.update(kwargs)

        start = time.time()
        response = requests.get(self.base_url, params=params)
        end = time.time()

        LOGGER.debug(
            'fetched amendements with search params: %s in %0.2f s',
            params,
            end - start
        )

        return parse_amendements_summary(response.url, response.json())

    def total_count(self, **kwargs):
        kwargs_copy = kwargs.copy()
        kwargs_copy['rows'] = 1
        response = self.get(**kwargs_copy)
        return response.total_count

    def iterator(self, **kwargs):
        rows = kwargs.get('rows', self.default_params['rows'])

        response = self.get(**kwargs)

        LOGGER.debug('start to fetch %s amendements with page size of %s',
                     response.total_count,
                     rows)
        LOGGER.debug('amendements fetched: %s / %s (%.1f%%)',
                     rows,
                     response.total_count,
                     rows / response.total_count * 100)

        yield response

        for start in range(rows, response.total_count, rows):
            LOGGER.debug('amendements fetched: %s / %s (%.1f%%)',
                         rows + start,
                         response.total_count,
                         (rows + start) / response.total_count * 100)
            kwargs_copy = kwargs.copy()
            kwargs_copy['start'] = start + 1
            yield self.get(**kwargs_copy)

    def get_order(self, **kwargs):
        iterator = AmendementSearchService().iterator(**kwargs)
        order = []
        for it in iterator:
            order += [amendement.num_amend for amendement in it.results]
        return order


@attr.s
class AmendementSearchResult(object):
    url = attr.ib(default=None)
    total_count = attr.ib(default=None)
    start = attr.ib(default=None)
    size = attr.ib(default=None)
    results = attr.ib(default=None)


@attr.s
class AmendementSummary(object):
    id = attr.ib(default=None)
    num_init = attr.ib(default=None)
    titre_dossier_legislatif = attr.ib(default=None)
    url_dossier_legislatif = attr.ib(default=None)
    instance = attr.ib(default=None)
    num_amend = attr.ib(default=None)
    url_amend = attr.ib(default=None)
    designation_article = attr.ib(default=None)
    designation_alinea = attr.ib(default=None)
    date_depot = attr.ib(default=None)
    signataires = attr.ib(default=None)
    sort = attr.ib(default=None)
    legislature = attr.ib(default=None)
    mission_visee = attr.ib(default=None)
    sous_reserve_de_traitement = attr.ib(default=None)


@attr.s
class Amendement(object):
    url = attr.ib(default=None)
    num_amtxt = attr.ib(default=None)
    amend_parent = attr.ib(default=None)
    url_dossier = attr.ib(default=None)
    num_init = attr.ib(default=None)
    etape = attr.ib(default=None)
    deliberation = attr.ib(default=None)
    titre_init = attr.ib(default=None)
    num_partie = attr.ib(default=None)
    designation_article = attr.ib(default=None)
    url_division = attr.ib(default=None)
    designation_alinea = attr.ib(default=None)
    mission = attr.ib(default=None)
    auteurs = attr.ib(default=None)
    auteur_id = attr.ib(default=None)
    groupe_id = attr.ib(default=None)
    cosignataires_id = attr.ib(default=None)
    seance = attr.ib(default=None)
    sort = attr.ib(default=None)
    date_badage = attr.ib(default=None)
    date_sort = attr.ib(default=None)
    ordre_texte = attr.ib(default=None)
    code = attr.ib(default=None)
    refcode = attr.ib(default=None)
    legislature = attr.ib(default=None)
    dispositif = attr.ib(default=None)
    expose = attr.ib(default=None)
    num_amend = attr.ib(default=None)

    @staticmethod
    def download_and_build(url):
        return parse_amendement(url, requests.get(url).content)


def clean_text(text):
    return text.strip().replace('\n', '').replace(u'\u2019', '\'')


def remove_inline_css_and_invalid_tags(soup):
    if soup is None:
        return u''

    if soup.div:
        soup = soup.div

    for invalid_tag in ['b', 'i', 'u']:
        for match in soup.findAll(invalid_tag):
            match.unwrap()

    # remove comments
    [comment.extract() for comment in
        soup.findAll(text=lambda text: isinstance(text, Comment))]

    for tag in soup:
        if type(tag) != NavigableString:
            del tag["style"]
            del tag["class"]

    return soup.decode_contents()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
from builtins import filter, str

import requests
from bs4 import BeautifulSoup

from .utils import extract_datetime


class Scrutin(object):
    def __init__(self, url=None, title=None, legislature=None, numero=None,
                 date=None, synthese=None, groupes=[]):
        self.url = url
        self.title = title
        self.legislature = legislature
        self.numero = numero
        self.date = date
        self.synthese = synthese
        self.groupes = groupes

    @staticmethod
    def download_and_build(url):
        return ScrutinParser(url, requests.get(url).content).parse()

    def to_dict(self):
        return {
            'url': self.url,
            'title': self.title,
            'legislature': self.legislature,
            'numero': self.numero,
            'date': self.date,
            'synthese': self.synthese.to_dict() if self.synthese else {},
            'groupes': {g.groupe: g.to_dict() for g in self.groupes}
        }


class ScrutinSynthese(object):
    ADOPTE = 'adopté'
    REJETE = 'rejeté'

    def __init__(self, votants=None, exprimes=None, majorite_absolue=None,
                 pour=None, contre=None, resultat=None):
        self.votants = votants
        self.exprimes = exprimes
        self.majorite_absolue = majorite_absolue
        self.pour = pour
        self.contre = contre
        self.resultat = resultat

    @staticmethod
    def build(url, html):
        return ScrutinSyntheseParser(url, html).parse()

    def to_dict(self):
        return {
            'votants': self.votants,
            'exprimes': self.exprimes,
            'majorite_absolue': self.majorite_absolue,
            'pour': self.pour,
            'contre': self.contre,
            'resultat': self.resultat
        }


class ScrutinGroupe(object):
    VOTE_POUR = 'pour'
    VOTE_CONTRE = 'contre'
    VOTE_ABSTENTION = 'abstention'
    VOTE_NON_VOTANT = 'non-votant'

    def __init__(self, groupe=None, pour=[], contre=[], abstention=[],
                 nonvotants=[]):
        self.groupe = groupe
        self.pour = pour
        self.contre = contre
        self.abstention = abstention
        self.nonvotants = nonvotants

    @staticmethod
    def build(url, soup):
        return ScrutinGroupeParser(url, soup).parse()

    def to_dict(self):
        return {
            'pour': self.pour,
            'contre': self.contre,
            'abstention': self.abstention,
            'non-votants': self.nonvotants
        }


class ScrutinParser(object):
    RE_DATE = re.compile(r'(\d+/\d+/\d+)')
    RE_SCRUTIN_URL = re.compile(
        r'/scrutins/detail/\(legislature\)/(\d+)/\(num\)/(\d+)')

    def __init__(self, url, html):
        self.url = url
        self.html = html
        self.soup = BeautifulSoup(html, 'html5lib')

    def parse(self):
        return Scrutin(
            url=self.url,
            title=self.parse_title(),
            legislature=self.parse_legislature(),
            numero=self.parse_numero(),
            date=self.parse_date(),
            synthese=ScrutinSynthese.build(self.url, self.html),
            groupes=self.parse_groupes()
        )

    def parse_title(self):
        return self.soup.h3.text if self.soup.h3 else None

    def parse_legislature(self):
        m = self.RE_SCRUTIN_URL.search(self.url)
        if not m:
            return None

        return int(m.group(1))

    def parse_numero(self):
        m = self.RE_SCRUTIN_URL.search(self.url)
        if not m:
            return None

        return int(m.group(2))

    def parse_date(self):
        if not self.soup.title:
            return None

        m = self.RE_DATE.search(self.soup.title.text)
        return extract_datetime(m.group(1)) if m else None

    def parse_groupes(self):
        return [
            ScrutinGroupe.build(self.url, tt)
            for tt in self.soup.select('.TTgroupe')
        ]


class ScrutinSyntheseParser(object):
    def __init__(self, url, html):
        self.url = url
        self.html = html

        self.soup = BeautifulSoup(html, 'html5lib')

    def parse(self):
        return ScrutinSynthese(
            votants=self.parse_votants(),
            exprimes=self.parse_exprimes(),
            majorite_absolue=self.parse_majorite_absolue(),
            pour=self.parse_pour(),
            contre=self.parse_contre(),
            resultat=self.parse_resultat()
        )

    def get_number(self, soup):
        return int(soup.b.text) if soup.b else None

    def get_number_by_id(self, id):
        def match(node):
            return node.has_attr('id') and node['id'] == id

        return self.get_number(
            next(filter(match, self.soup.select('.repartitionvotes'))))

    def get_number_by_text(self, text):
        def match(node):
            return node.text.startswith(text)

        return self.get_number(
            next(filter(match, self.soup.select('.repartitionvotes'))))

    def parse_votants(self):
        return self.get_number_by_id('total')

    def parse_exprimes(self):
        return self.get_number_by_text('Nombre de suffrages exprimés')

    def parse_majorite_absolue(self):
        return self.get_number_by_text('Majorité absolue')

    def parse_pour(self):
        return self.get_number_by_id('pour')

    def parse_contre(self):
        return self.get_number_by_id('contre')

    def parse_resultat(self):
        node = self.soup.select('.annonce .annoncevote')[0]
        if not node:
            return None

        return (ScrutinSynthese.ADOPTE
                if ScrutinSynthese.ADOPTE in node.text
                else ScrutinSynthese.REJETE)


class ScrutinGroupeParser(object):
    RE_NOM = re.compile(r'(.*) \(\d+ membres?\)')
    RE_DEPUTE = re.compile(r'([^> ]+(?: d(?:e|u|es))?) <b>([^<]+)</b>')

    def __init__(self, url, soup):
        self.url = url
        self.soup = soup

    def parse(self):
        return ScrutinGroupe(
            groupe=self.parse_groupe(),
            pour=self.parse_pour(),
            contre=self.parse_contre(),
            abstention=self.parse_abstentions(),
            nonvotants=self.parse_nonvotants()
        )

    def parse_groupe(self):
        node = self.soup.select('p.nomgroupe')
        if not len(node):
            return None

        m = self.RE_NOM.match(node[0].text)
        return m.group(1) if m else None

    def parse_liste_deputes(self, selector):
        ul = self.soup.select(selector)

        if len(ul):
            # Parsing html direct car dans certains cas le <ul class="deputes">
            # ne contient que du texte, pas de <li>.
            html = str(ul[0]).replace('\xa0', ' ')
            return [
                '%s %s' % nom
                for nom in self.RE_DEPUTE.findall(html)
            ]

        return []

    def parse_pour(self):
        return self.parse_liste_deputes('.Pour ul.deputes')

    def parse_contre(self):
        return self.parse_liste_deputes('.Contre ul.deputes')

    def parse_abstentions(self):
        return self.parse_liste_deputes('.Abstention ul.deputes')

    def parse_nonvotants(self):
        return self.parse_liste_deputes('.Non-votants ul.deputes')



# -*- coding: utf-8 -*-
import re

import attr
import requests
import xmltodict
from bs4 import BeautifulSoup


def parse_question(url, xml):
    data = xmltodict.parse(xml)['QUESTION']
    data['URL'] = url
    return data


def parse_question_search_result(url, html_content):
    soup = BeautifulSoup(html_content, "html5lib")

    search_result = QuestionSearchResult(**{
        'url': url,
        'total_count': int(soup.find('article').div.div.p.strong.text)
    })

    results = []
    for tr in soup.find_all('tr')[1:]:
        all_tds = tr.find_all('td')
        url = all_tds[0].a['href']
        legislature, numero, question_type = re.search(
            r'(\d+)-(\d+)(QE|QOSD)\.htm', url).groups()
        dates = all_tds[2].find_all('strong')
        results.append(QuestionSummary(**{
            'url': url,
            'legislature': legislature,
            'numero': numero,
            'question_type': question_type,
            'author': all_tds[1].strong.text,
            'tags': all_tds[1].em.text,
            'publication_date': dates[0].text,
            'answer_date': dates[1].text if len(dates) > 1 else None,
        }))

    search_result.results = results
    next_link = soup.find('div',
                          class_='pagination-bootstrap').find_all('li')[-1]
    search_result.next_url = next_link.a['href'] if next_link.a else None

    return search_result


class QuestionSearchService(object):
    def __init__(self):
        self.base_url = 'http://www2.assemblee-nationale.fr/'
        self.search_url = '%srecherche/resultats_questions' % self.base_url
        self.default_params = {
            'limit': 10,
            'legislature': None,
            'replies[]': None,  # ar, sr
            'removed[]': None,  # 0,1
            'ssTypeDocument[]': 'qe',
        }

    def get(self, legislature=14, is_answered=None, is_removed=None, size=10):
        params = self.default_params.copy()

        if is_answered:
            is_answered = 'ar'
        elif is_answered is not None:
            is_answered = 'sr'
        if is_removed is not None:
            is_removed = int(is_removed)

        params.update({
            'legislature': legislature,
            'limit': size,
            'replies[]': is_answered,
            'removed[]': is_removed
        })
        response = requests.post(self.search_url, data=params)

        return parse_question_search_result(response.url, response.content)

    def total_count(self, legislature=14, is_answered=None, is_removed=None):
        return self.get(legislature=legislature, is_answered=is_answered,
                        is_removed=is_removed, size=1).total_count

    def iter(self, legislature=14, is_answered=None, is_removed=None, size=10):
        search_results = self.get(legislature=legislature,
                                  is_answered=is_answered,
                                  is_removed=is_removed, size=size)
        yield search_results

        for start in range(1, search_results.total_count, size):
            if search_results.next_url is not None:
                yield parse_question_search_result(
                    search_results.next_url,
                    requests.get(self.base_url +
                                 search_results.next_url).content)


@attr.s
class QuestionSearchResult(object):
    url = attr.ib(default=None)
    total_count = attr.ib(default=None)
    next_url = attr.ib(default=None)
    size = attr.ib(default=None)
    results = attr.ib(default=None)


@attr.s
class QuestionSummary(object):
    url = attr.ib(default=None)
    legislature = attr.ib(default=None)
    numero = attr.ib(default=None)
    question_type = attr.ib(default=None)
    author = attr.ib(default=None)
    tags = attr.ib(default=None)
    publication_date = attr.ib(default=None)
    answer_date = attr.ib(default=None)



# -*- coding: utf-8 -*-
from builtins import str
import json
from datetime import datetime

import dateparser
import re

hours_with_minutes_re = re.compile(' heures ')
hours_without_minutes = re.compile(' heures$')


def extract_datetime(text):
    text = hours_with_minutes_re.sub(':', text.strip())
    text = hours_without_minutes.sub(':00', text)

    return dateparser.parse(text, languages=['fr'])


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


class JSONDecoder(json.JSONDecoder):
    DATETIME_PATTERN = re.compile(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}")

    def __init__(self, *args, **kwargs):
        super(JSONDecoder, self).__init__(
            object_pairs_hook=self.object_pairs_hook,
            *args,
            **kwargs)

    def object_pairs_hook(self, obj):
        return dict((k, self.decode_on_match(v)) for k, v in obj)

    def decode_on_match(self, obj):
        string = str(obj)

        match = re.search(self.DATETIME_PATTERN, string)
        if match:
            return datetime.strptime(match.string, '%Y-%m-%dT%H:%M:%S')

        return obj


def json_dumps(data, *args, **kwargs):
    return JSONEncoder(*args, **kwargs).encode(data)


def json_loads(string, *args, **kwargs):
    return JSONDecoder(*args, **kwargs).decode(string)


def convert_camelcase_to_underscore(name):
    # thx to http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-camel-case  # noqa
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



# -*- coding: utf-8 -*-

import re
from setuptools import setup
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

with open('anpy/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

with open(path.join(here, 'README.md'), encoding='utf-8') as readme:
    LONG_DESC = readme.read()

with open(path.join(here, 'requirements.txt')) as f:
    requirements = f.read().splitlines()

setup(
    name='anpy',
    version=version,
    description='Python client for assemblee-nationale.fr website',
    long_description=LONG_DESC,
    long_description_content_type='text/markdown',
    license="MIT",

    url='https://github.com/regardscitoyens/anpy',
    author='Regards Citoyens',
    author_email='contact@regardscitoyens.org',

    include_package_data=True,

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],

    keywords='scraping politics data',

    packages=['anpy'],

    install_requires=requirements,

    scripts=['bin/anpy-cli'],
)



"""
Experiment to use the JSON formatted data provided by the AN for the doslegs

  python dossier_from_opendata.py <dosleg_url>
"""

import sys
import json
import zipfile
import io
import re

from lawfactory_utils.urls import download, enable_requests_cache, clean_url, parse_national_assembly_url


OPENDATA_ID_REGEX = r'(.{4})([ANS]*)(R[0-9])([LS]*)([0-9]*)([BTACP]*)(.*)'


def same_stage_step_instit(a, b):
    return a.get('stage') == b.get('stage') and a.get('step') == b.get('step') \
        and a.get('institution') == b.get('institution')


def yield_leafs(etape, path=None):
    if path is None:
        path = []
    if etape.get("actesLegislatifs"):
        for children in to_arr(etape["actesLegislatifs"]["acteLegislatif"]):
            yield from yield_leafs(children, path=path + [etape])
    else:
        yield path, etape


# TODO : make this generic for the latest legislature, for instance by finding the url of the zip in http://data.assemblee-nationale.fr/reunions/reunions
def find_texts_discussed_after(min_date, senate_urls=False, include_resolutions=False):
    OPEN_DATA_REUNIONS_URL = "http://data.assemblee-nationale.fr/static/openData/repository/15/vp/reunions/Agenda_XV.json.zip"
    reunions = convert_reunions_open_data_file(download_open_data_file(None, OPEN_DATA_REUNIONS_URL))

    doslegs = set()
    for reunion in to_arr(reunions['reunions']['reunion']):

        date = reunion['timeStampDebut'].split('T')[0]
        if date < min_date:
            continue

        if not reunion.get('ODJ') or not reunion['ODJ'].get('pointsODJ'):
            continue

        for pointODJ in to_arr(reunion['ODJ']['pointsODJ']['pointODJ']):
            if not pointODJ['dossiersLegislatifsRefs']:
                continue
            for dosleg in to_arr(pointODJ['dossiersLegislatifsRefs']['dossierRef']):
                doslegs.add(dosleg)

    dossiers_json = download_open_data_doslegs(15)
    docs = {doc['dossierParlementaire']["uid"]: doc['dossierParlementaire'] for doc in dossiers_json["export"]["dossiersLegislatifs"]["dossier"]}

    doslegs_urls = set()
    for dosleg_ref in doslegs:
        if dosleg_ref not in docs:
            # TODO: dossierAbsorbantRef
            print('[anpy]', dosleg_ref, ' dosleg in ODJ but not found in Open Data', file=sys.stderr)
            continue
        dossier = docs[dosleg_ref]

        if dossier["@xsi:type"] != "DossierLegislatif_Type":
            continue

        if not include_resolutions and dossier["procedureParlementaire"]["libelle"] == "Résolution":
            continue

        titreChemin = dossier["titreDossier"]["titreChemin"]
        url_pattern = "http://www.assemblee-nationale.fr/dyn/{}/dossiers/{}"
        url = url_pattern.format(dossier["legislature"], titreChemin)
        url_senat = dossier["titreDossier"]["senatChemin"]
        if url_senat and senate_urls:
            url = clean_url(url_senat)
        doslegs_urls.add(url)

    return doslegs_urls


def to_arr(obj):
    if type(obj) is list:
        return obj
    return [obj]


def test_status(url):
    try:
        resp = download(url)
        if resp.status_code != 200:
            return False
    except Exception:
        return False
    return resp


def download_open_data_file(filename, file_url):
    raw_data = download(file_url)
    data_zip = zipfile.ZipFile(io.BytesIO(raw_data.content))
    if filename:
        with data_zip.open(filename) as d:
            return json.loads(d.read().decode('utf-8'))
    return data_zip


def convert_dossiers_open_data_file(data_zip):
    data = {
        "export": {
            "@xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "dossiersLegislatifs": {
                "dossier": []
            },
            "textesLegislatifs": {
                "document": []
            }
        }
    }
    for filename in data_zip.namelist():
        with data_zip.open(filename) as d:
            filedata = json.loads(d.read().decode('utf-8'))
            if "dossierParlementaire" in filename:
                data["export"]["dossiersLegislatifs"]["dossier"].append(filedata)
            else:
                data["export"]["textesLegislatifs"]["document"].append(filedata["document"])
    return data


def convert_reunions_open_data_file(data_zip):
    data = {
        "reunions": {
            "reunion": []
        }
    }
    for filename in data_zip.namelist():
        with data_zip.open(filename) as d:
            filedata = json.loads(d.read().decode('utf-8'))
            data["reunions"]["reunion"].append(filedata["reunion"])
    return data


def download_open_data_doslegs(legislature):
    files = {
        15: (
            None,
            "http://data.assemblee-nationale.fr/static/openData/repository/15/loi/dossiers_legislatifs/Dossiers_Legislatifs_XV.json.zip",
        ),
        14: (
            "Dossiers_Legislatifs_XIV.json",
            "http://data.assemblee-nationale.fr/static/openData/repository/14/loi/dossiers_legislatifs/Dossiers_Legislatifs_XIV.json.zip",
        ),
    }
    filename, file_url = files[legislature]
    data = download_open_data_file(filename, file_url)
    if filename is None:
        data = convert_dossiers_open_data_file(data)
    return data


def an_text_url_opendata(uid):
    return "http://www.assemblee-nationale.fr/dyn/opendata/%s.html" % uid


def an_text_url(identifiant, code):
    """
    Port of the PHP function used by the National Assembly:

    public function urlOpaque($identifiant, $codeType = NULL)
    {
        $datas = array(
            'PRJL' => array('repertoire' => 'projets', 'prefixe' => 'pl', 'suffixe' => ''),
            'PION' => array('repertoire' => 'propositions', 'prefixe' => 'pion', 'suffixe' => ''),
            'PNRECOMENQ' => array('repertoire' => 'propositions', 'prefixe' => 'pion', 'suffixe' => ''),
            'PNREAPPART341' => array('repertoire' => 'propositions', 'prefixe' => 'pion', 'suffixe' => ''),
            'PNREMODREGLTAN' => array('repertoire' => 'propositions', 'prefixe' => 'pion', 'suffixe' => ''),
            'AVCE' => array('repertoire' => 'projets', 'prefixe' => 'pl', 'suffixe' => '-ace'),
            'ETDI' => array('repertoire' => 'projets', 'prefixe' => 'pl', 'suffixe' => '-ei'),
            'ACIN' => array('repertoire' => 'projets', 'prefixe' => 'pl', 'suffixe' => '-ai'),
            'LETT' => array('repertoire' => 'projets', 'prefixe' => 'pl', 'suffixe' => '-l'),
            'PNRETVXINSTITEUROP' => array('repertoire' => 'europe/resolutions', 'prefixe' => 'ppe', 'suffixe' => ''),
            'PNRE' => array('repertoire' => 'europe/resolutions', 'prefixe' => 'ppe', 'suffixe' => ''),
            'RION' => array('repertoire' => '', 'prefixe' => '', 'suffixe' => ''),
            'TCOM' => array('repertoire' => 'ta-commission', 'prefixe' => 'r', 'suffixe' => '-a0'),
            'TCOMMODREGLTAN' => array('repertoire' => 'ta-commission', 'prefixe' => 'r', 'suffixe' => '-a0'),
            'TCOMTVXINSTITEUROP' => array('repertoire' => 'ta-commission', 'prefixe' => 'r', 'suffixe' => '-a0'),
            'TCOMCOMENQ' => array('repertoire' => 'ta-commission', 'prefixe' => 'r', 'suffixe' => '-a0'),
            'TADO' => array('repertoire' => 'ta', 'prefixe' => 'ta', 'suffixe' => ''),
        );
        preg_match('/(.{4})([ANS]*)(R[0-9])([LS]*)([0-9]*)([BTACP]*)(.*)/', $identifiant, $matches);
        $leg = $matches[5];
        $typeTa = $matches[6];
        $num = $matches[7];
        switch ($typeTa) {
            case 'BTC':
                $type = 'TCOM';
                break;
            case 'BTA':
                $type = 'TADO';
                break;
            default:
                $type = $codeType;
        }
        $host = "http://www.assemblee-nationale.fr/";
        return $host . $leg . "/" . $datas[$type]['repertoire'] . "/" . $datas[$type]['prefixe'] . $num . $datas[$type]['suffixe'] . ".pdf";
    }
    """
    datas = {
        'PRJL': {
            'repertoire': 'projets',
            'prefixe': 'pl',
            'suffixe': '',
        },
        'PION': {
            'repertoire': 'propositions',
            'prefixe': 'pion',
            'suffixe': '',
        },
        'PNRECOMENQ': {
            'repertoire': 'propositions',
            'prefixe': 'pion',
            'suffixe': '',
        },
        'PNREAPPART341': {
            'repertoire': 'propositions',
            'prefixe': 'pion',
            'suffixe': '',
        },
        'PNREMODREGLTAN': {
            'repertoire': 'propositions',
            'prefixe': 'pion',
            'suffixe': '',
        },
        'AVCE': {
            'repertoire': 'projets',
            'prefixe': 'pl',
            'suffixe': '-ace',
        },
        'ETDI': {
            'repertoire': 'projets',
            'prefixe': 'pl',
            'suffixe': '-ei',
        },
        'ACIN': {
            'repertoire': 'projets',
            'prefixe': 'pl',
            'suffixe': '-ai',
        },
        'LETT': {
            'repertoire': 'projets',
            'prefixe': 'pl',
            'suffixe': '-l',
        },
        'PNRETVXINSTITEUROP': {
            'repertoire': 'europe/resolutions',
            'prefixe': 'ppe',
            'suffixe': '',
        },
        'PNRE': {
            'repertoire': 'propositions',
            'prefixe': 'pion',
            'suffixe': '',
        },
        'RION': {
            'repertoire': '',
            'prefixe': '',
            'suffixe': '',
        },
        'TCOM': {
            'repertoire': 'ta-commission',
            'prefixe': 'r',
            'suffixe': '-a0',
        },
        'TCOMMODREGLTAN': {
            'repertoire': 'ta-commission',
            'prefixe': 'r',
            'suffixe': '-a0',
        },
        'TCOMTVXINSTITEUROP': {
            'repertoire': 'ta-commission',
            'prefixe': 'r',
            'suffixe': '-a0',
        },
        'TCOMCOMENQ': {
            'repertoire': 'ta-commission',
            'prefixe': 'r',
            'suffixe': '-a0',
        },
        'TADO': {
            'repertoire': 'ta',
            'prefixe': 'ta',
            'suffixe': '',
        },
        # NOT IN NATIONAL ASSEMBLY PHP CODE
        'RAPP': {
            'repertoire': 'rapports',
            'prefixe': 'r',
            'suffixe': '',
        },
        'RINF': {
            'repertoire': 'rap-inf',
            'prefixe': 'i',
            'suffixe': '',
        }
    }
    match = re.match(OPENDATA_ID_REGEX, identifiant)
    leg = match.group(5)
    typeTa = match.group(6)
    num = match.group(7)
    if typeTa == 'BTC':
        type = 'TCOM'
    elif typeTa == 'BTA':
        type = 'TADO'
    elif typeTa == 'TAP':
        type = 'TADO'
    else:
        type = code
    host = "http://www.assemblee-nationale.fr/"

    if type not in datas:
        # ex: ALCNANR5L15B0002 (allocution du président)
        raise Exception('Unknown document type for %s' % identifiant)

    return host + leg + "/" + datas[type]['repertoire'] + "/" + datas[type]['prefixe'] + num + datas[type]['suffixe'] + ".asp"


def parse(url, logfile=sys.stderr, cached_opendata_an={}):
    def _log(*args):
        nonlocal logfile
        print(*args, file=logfile)

    legislature_parsed, _ = parse_national_assembly_url(url)

    # try to find the dosleg in current and past legislature
    for legislature in (legislature_parsed, legislature_parsed - 1):
        if legislature and legislature in cached_opendata_an:
            dossiers_json = cached_opendata_an[legislature]
        else:
            dossiers_json = download_open_data_doslegs(legislature)

        docs = {doc["uid"]: doc for doc in dossiers_json["export"]["textesLegislatifs"]["document"]}

        for dossier in dossiers_json["export"]["dossiersLegislatifs"]["dossier"]:
            dossier = dossier["dossierParlementaire"]

            if dossier.get("@xsi:type") != "DossierLegislatif_Type":
                continue

            titreChemin = dossier["titreDossier"]["titreChemin"]

            # find the right dosleg even if it's an old url
            url_common_part = "{}/dossiers/{}".format(dossier["legislature"], titreChemin)
            if not url.endswith(url_common_part):
                # dosleg might be from previous legislature
                url_common_part = "{}/dossiers/{}".format(int(dossier["legislature"]) + 1, titreChemin)
                if not url.endswith(url_common_part):
                    continue
            url = "http://www.assemblee-nationale.fr/dyn/{}".format(url_common_part)

            data = {}
            data["urgence"] = False
            url_senat = dossier["titreDossier"]["senatChemin"]
            if url_senat:
                data["url_dossier_senat"] = clean_url(url_senat)
            data["long_title"] = dossier["titreDossier"]["titre"]
            data["url_dossier_assemblee"] = clean_url(url)
            data["assemblee_legislature"] = int(dossier["legislature"])
            data["assemblee_slug"] = dossier["titreDossier"]["titreChemin"]
            data["assemblee_id"] = "%s-%s" % (dossier["legislature"], data["assemblee_slug"])

            if dossier["procedureParlementaire"]["libelle"] in (
                "Projet de loi de finances de l'année",
                "Projet de loi de financement de la sécurité sociale",
                "Projet de loi de finances rectificative",
                "Projet ou proposition de loi constitutionnelle",
            ):
                data['use_old_procedure'] = True

            data["steps"] = []
            step = None
            start_step = None
            for etape in to_arr(dossier["actesLegislatifs"]["acteLegislatif"]):
                for path, sous_etape in yield_leafs(etape):
                    if sous_etape["@xsi:type"] in ("EtudeImpact_Type", "DepotAvisConseilEtat_Type"):
                        continue

                    step = {}

                    date = sous_etape.get("dateActe")
                    if date:
                        step["date"] = date.split("T")[0]

                    if sous_etape["@xsi:type"] == "ProcedureAccelere_Type":
                        data["urgence"] = True
                        continue
                    elif sous_etape["@xsi:type"] == "Promulgation_Type":
                        url = clean_url(sous_etape.get("urlLegifrance") or sous_etape["infoJO"]["urlLegifrance"])
                        data["url_jo"] = url
                        data["end"] = step["date"]

                        step["institution"] = "gouvernement"
                        step["stage"] = "promulgation"
                        step["source_url"] = url
                        data["steps"].append(step)
                        continue
                    elif sous_etape["@xsi:type"] == "ConclusionEtapeCC_Type":
                        step["institution"] = "conseil constitutionnel"
                        step["stage"] = "constitutionnalité"
                        step["source_url"] = clean_url(sous_etape["urlConclusion"])
                        data["steps"].append(step)

                    if "textesAssocies" in sous_etape:
                        # TODO review
                        sous_etape["texteAssocie"] = to_arr(sous_etape["textesAssocies"]["texteAssocie"])[0]["refTexteAssocie"]

                    code = sous_etape.get("codeActe")

                    if "AVIS-RAPPORT" in code or code == 'CMP-DEPOT':
                        continue
                    if '-DPTLETTRECT' in code:
                        continue

                    if code.startswith("AN"):
                        step["institution"] = "assemblee"
                    elif code.startswith("SN"):
                        step["institution"] = "senat"

                    if "-DEPOT" in code:
                        step["step"] = "depot"
                    elif "-COM" in code:
                        step["step"] = "commission"
                    elif "-DEBATS" in code:
                        step["step"] = "hemicycle"
                    else:
                        _log("  - WARNING Unknown step type", code)
                        continue

                    if "1-" in code:
                        step["stage"] = "1ère lecture"
                    elif "2-" in code:
                        step["stage"] = "2ème lecture"
                    elif "3-" in code:
                        step["stage"] = "3ème lecture"  # TODO: else libelleCourt
                    elif "NLEC-" in code:
                        step["stage"] = "nouv. lect."
                    elif "ANLDEF-" in code:
                        step["stage"] = "l. définitive"
                        if step["step"] == "commission":
                            continue
                    elif "CMP-" in code:
                        step["stage"] = "CMP"
                        if "-DEBATS-AN" in code:
                            step["institution"] = "assemblee"
                        elif "-SN" in code:
                            step["institution"] = "senat"
                            if "RAPPORT-SN" in code:
                                # ignore the cmp_commission_other_url for now
                                continue
                        else:
                            step["institution"] = "CMP"
                    elif "ANLUNI-" in code:
                        step["stage"] = "l. unique"

                    step["id_opendata"] = sous_etape["uid"]

                    # keep first step for a step-type (ex: first hemiycle)
                    if start_step is None or not same_stage_step_instit(start_step, step):
                        start_step = step

                    if "texteAdopte" in sous_etape or "texteAssocie" in sous_etape:
                        # there is no multiple depot in the National Assembly
                        # simply the senate re-submitting the same text
                        if data['steps']:
                            last_step = data['steps'][-1]
                            if last_step['institution'] == 'assemblee' and last_step.get('step') == step.get('step') == 'depot':
                                # ignore the depot we already have (since the new one is the same)
                                data['steps'] = data['steps'][:-1]

                        # step['xsi-type'] = sous_etape.get('@xsi:type')
                        # step['code'] = sous_etape.get('codeActe')

                        id_text = sous_etape.get("texteAdopte") or sous_etape.get("texteAssocie")
                        if id_text:
                            if "proposal_type" not in data:
                                if id_text.startswith("PRJL"):
                                    data["proposal_type"] = "PJL"
                                elif id_text.startswith("PION"):
                                    data["proposal_type"] = "PPL"

                            doc = {}
                            if id_text in docs:
                                doc = docs[id_text]
                            else:
                                _log("  - ERROR missing text", id_text)

                            url = None
                            if step.get("institution") == "assemblee" or "-AN" in code:
                                doc_code = None
                                if doc:
                                    doc_code = doc['classification']['type']['code']
                                    if doc_code == 'ACIN':
                                        continue

                                match = re.match(OPENDATA_ID_REGEX, id_text)
                                step_legislature = int(match.group(5))
                                if step_legislature >= 15:
                                    url = an_text_url_opendata(id_text)
                                else:
                                    url = an_text_url(id_text, doc_code)
                                if url:
                                    step['source_url'] = url

                                    # hardfix and should be fixed by the AN one day
                                    if data["assemblee_id"] == '15-retablissement_confiance_action_publique':
                                        if step["stage"] == "l. définitive" and step["step"] == "depot":
                                            step['source_url'] = 'http://www.assemblee-nationale.fr/dyn/opendata/PRJLANR5L15BTA0017.html'

                        data["steps"].append(step)

                    else:
                        pass

            if data['steps']:
                # add predicted step
                if not data.get('url_jo'):
                    if data['steps'][-1].get('step') != start_step.get('step') and start_step.get('step'):
                        # TODO: we could also add all the dates into a steps['dates'] = [..]
                        data['steps'].append(start_step)
                data["beginning"] = data["steps"][0]["date"]
            else:
                _log("  - WARNING no steps found for", url)

            return data
    return []


if __name__ == "__main__":
    enable_requests_cache()
    url = sys.argv[1]
    data = parse(url)
    print(json.dumps(data, indent=2, sort_keys=True, ensure_ascii=False))
